"""
    audit permission
"""
from typing import Optional

from django.conf import settings
from django.contrib.auth.models import Permission
from django.contrib.contenttypes.models import ContentType
from django.db.models import Model


class PermissionSettings:
    """
    permission settings data object
    """

    def __init__(self):
        auditmatic_settings = settings.AUDITMATIC
        self.codename = "can_audit"
        self.name = "Can audit"
        permission = auditmatic_settings.get("permission")
        if not permission:
            return

        if permission_codename := permission.get("codename"):
            self.codename = permission_codename
        if permission_name := permission.get("name"):
            self.name = permission_name


permission_settings = PermissionSettings()


def get_audit_permission(model: Model) -> Optional[Permission]:
    """
        create an audit permission for the given model.
    :param model:
    :return:
    """
    content_type = ContentType.objects.get_for_model(model)
    print(content_type.id, content_type)
    lowered_model = model._meta.object_name.lower()
    try:
        return Permission.objects.get(
            codename=f"{permission_settings.codename}_{lowered_model}",
            name=f"{permission_settings.name} {lowered_model}",
            content_type=content_type,
        )
    except Permission.DoesNotExist as ex:
        print(ex)


def get_or_create_audit_permission(model: Model) -> Permission:
    """
        create an audit permission for the given model.
    :param model:
    :return:
    """
    content_type = ContentType.objects.get_for_model(model)
    lowered_model = model._meta.object_name.lower()
    obj, _ = Permission.objects.get_or_create(
        codename=f"{permission_settings.codename}_{lowered_model}",
        name=f"{permission_settings.name} {lowered_model}",
        content_type=content_type,
    )
    return obj



"""
    django command to install audit permissions
"""
from django.core.management.base import BaseCommand

from django_auditmatic.install.audit_permission import install_audit_permission


class Command(BaseCommand):
    """
    install_audit command
    """

    help = "Installs audit table and triggers for the configured models."

    def add_arguments(self, parser):
        # parser.add_argument('models', nargs='+', type=str)
        pass

    def handle(self, *args, **options):
        install_audit_permission()



from typing import List, Optional, Tuple

from django.apps import apps
from django.conf import settings


def find_schemas() -> Optional[List[str]]:
    """
        find all schemas by querying the tenant's schema names.
    :return: List[str]
    """
    if not hasattr(settings, "TENANT_MODEL"):
        return None

    tenant_model_setting = settings.TENANT_MODEL
    split_tenant_model = tenant_model_setting.split(".")
    tenant_model_name = split_tenant_model[-1]
    tenant_model = None
    for model in apps.get_models():
        # print(dir(model))
        # print(model._meta.model_name, tenant_model_setting)
        if model._meta.model_name == tenant_model_name:
            tenant_model = model
    if not tenant_model:
        return None
    return tenant_model.objects.values_list("schema_name", flat=True)


def get_tenant_schemas_and_apps() -> Tuple[List, List]:
    """
        process tenant configuration if tenants are in use.
    :return:
    """
    tenant_schemas = find_schemas()
    schema_apps = []
    if tenant_schemas and len(tenant_schemas) > 0:
        if not hasattr(settings, "TENANT_APPS"):
            raise RuntimeError(
                "Detected tenant model but no apps configured for TENANT_APPS setting."
            )
        schema_apps = settings.TENANT_APPS
    return tenant_schemas, schema_apps



"""
given steps
"""
from django.conf import settings
from django.contrib.auth.models import Group, User

from behave import given, use_step_matcher  # pylint: disable=E0611

from django_auditmatic.permission import get_or_create_audit_permission

use_step_matcher("parse")


@given("the user exists")
def user_exists(context):
    obj = User.objects.create_user(
        email="user@example.com", username="user", password="hellow", is_active=True
    )
    context.user = obj
    try:
        any_group = Group.objects.get(name="Any")
    except Group.DoesNotExist:
        any_group = Group.objects.create(name="Any")
    any_group.user_set.add(obj)
    assert context.user is not None


@given("the user model is configured with allow: any")
def user_model_configured_with_allow_any(context):
    """
    :type context: behave.runner.Context
    """
    settings.AUDITMATIC = {
        "apps": {
            "auth": {
                "User": {"allow": any},
            }
        }
    }


@given("the user model is configured")
def user_model_is_configured(context):
    """
    :type context: behave.runner.Context
    """
    settings.AUDITMATIC = {"apps": {"auth": {"User"}}}
    print(settings.AUDITMATIC)


@given("the user has the audit permission")
def user_has_audit_permission(context):
    """
    :type context: behave.runner.Context
    """
    permission = get_or_create_audit_permission(User)
    context.user.user_permissions.add(permission)


@given("the user does not have the audit permission")
def user_does_not_have_audit_permission(context):
    """
    :type context: behave.runner.Context
    """
    assert not context.user.has_perm("auth.can_audit_user")


@given("the audit name is {audit_name}")
def set_audit_name(context, audit_name: str):
    """
    :type context: behave.runner.Context
    :type audit_name: str
    """
    context.audit_name = audit_name


@given("the table name is {table_name}")
def set_table_name(context, table_name: str):
    """
    :type context: behave.runner.Context
    :type table_name: str
    """
    context.table_name = table_name



"""
    function to generate sql trigger
"""


def generate_trigger(audit_name, table_name, action):
    """
        generates create trigger sql statement
    :param audit_name:
    :param table_name:
    :param action:
    :return: create trigger statement
    """
    trigger_audit_name = audit_name.replace(".", "_")
    trigger_audit_name = f"{trigger_audit_name}_{action}"
    return f"""
    CREATE TRIGGER {trigger_audit_name}
        AFTER {action} ON {table_name}
            FOR EACH ROW
    EXECUTE PROCEDURE {audit_name}();
    """



"""
then steps
"""
from django.contrib.auth.models import User

from behave import then, use_step_matcher  # pylint: disable=E0611

from django_auditmatic.permission import get_audit_permission

use_step_matcher("parse")


@then("the statement generated should contain the audit name")
def generated_statement_contains_audit_name(context):
    """
    :type context: behave.runner.Context
    """
    context.test.assertTrue(context.audit_name in context.statement_generated)


@then("the statement generated should contain the table name")
def generated_statement_contains_table_name(context):
    """
    :type context: behave.runner.Context
    """
    context.test.assertTrue(context.audit_name in context.statement_generated)


@then("the statement generated should contain hstore")
def generated_statement_contains_hstore(context):
    """
    :type context: behave.runner.Context
    """
    context.test.assertTrue("hstore" in context.statement_generated)


@then("result is True")
def result_is_true(context):
    """
    :type context: behave.runner.Context
    """
    assert context.result


@then("the user model has an audit permission")
def user_model_has_audit_permission(context):
    """
    :type context: behave.runner.Context
    """
    permission = get_audit_permission(User)
    assert permission is not None


@then("result is False")
def result_is_false(context):
    """
    :type context: behave.runner.Context
    """
    assert not bool(context.result)



"""
    database utility functions
"""
from typing import List, Optional, Tuple

from django.db import connection
from django.db.models import Model

from django_auditmatic.configuration.data_classes import ConfiguredNames, ModelNames
from django_auditmatic.utils.generate import generate_install_hstore, generate_sql


def get_model_names(
    model: Model, configured_names: ConfiguredNames
) -> Optional[ModelNames]:
    """
        process the model for all configured schemas.
    :param model:
    :param configured_names:
    :return:
    """
    model_names = ModelNames.from_model(model)
    if model_names.app_name not in configured_names.app_names:
        return
    if model_names.model_name not in configured_names.model_names[model_names.app_name]:
        return
    return model_names


def process_model_for_all_schemas(
    model: Model,
    configured_names: ConfiguredNames,
    schema_apps,
    tenant_schemas,
):
    """
        process the model for all configured schemas.
    :param model:
    :param configured_names:
    :param schema_apps:
    :param tenant_schemas:
    :return:
    """
    model_names = get_model_names(model, configured_names)

    schema = "public"

    with connection.cursor() as cursor:
        cursor.execute(generate_install_hstore())
        if not len(schema_apps):  # pylint: disable=C1802
            process_model(cursor, configured_names.model_m2m_names, model_names, schema)
            return

        if model_names.app_name not in schema_apps:
            process_model(cursor, configured_names.model_m2m_names, model_names, schema)
            return

        for tenant_schema in tenant_schemas:
            process_model(
                cursor, configured_names.model_m2m_names, model_names, tenant_schema
            )


def process_model(cursor, configured_model_m2m_names, model_names, schema):
    """
        generates sql for the model and any many to many models configured.
    :param cursor:
    :param configured_model_m2m_names:
    :param model_names:
    :param schema:
    :param model:
    :return:
    """
    app_name = model_names.app_name
    model_name = model_names.model_name
    generate_and_execute(cursor, app_name, model_name, schema)
    m2m_key = f"{app_name}_{model_name}"
    is_any, m2m_names = get_m2m_names(configured_model_m2m_names, m2m_key)
    for field in model_names.model._meta.many_to_many:
        process_m2m_field(field, is_any, m2m_names, app_name, schema, cursor)


def generate_and_execute(
    cursor, app_name, model_name, schema, table_name: Optional[str] = None
):
    statement = generate_sql(app_name, model_name, schema, table_name=table_name)
    cursor.execute(statement)


def check_m2m_configured(field, m2m_names) -> bool:
    model_name = field.model._meta.model_name
    related_model_name = field.related_model._meta.model_name
    if (model_name, related_model_name) not in m2m_names:
        return False
    return True


def process_m2m_field(field, is_any, m2m_names, app_name, schema, cursor):
    name = field.m2m_db_table()
    if not is_any and not check_m2m_configured(field, m2m_names):
        return
    generate_and_execute(cursor, app_name, name, schema, table_name=name)


def get_m2m_names(configured_model_m2m_names, m2m_key) -> Tuple[List[str], bool]:
    """

    :param configured_model_m2m_names:
    :param m2m_key:
    :return: List of many-to-many names, use any name.
    """
    m2m_names = configured_model_m2m_names[m2m_key]
    if m2m_names == any or any in m2m_names:  # pylint: disable=W0143
        return [], True
    for m2m_name in m2m_names:
        m2m_names.append(
            (
                m2m_name[0].lower(),
                m2m_name[1].lower(),
            )
        )
    return m2m_names, False



"""
Django settings for django_model_auditmatic project.

Generated by 'django-admin startproject' using Django 4.0.2.

For more information on this file, see
https://docs.djangoproject.com/en/4.0/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/4.0/ref/settings/
"""

from pathlib import Path

import environ

env = environ.Env()
# Build paths inside the project like this: BASE_DIR / 'subdir'.
BASE_DIR = Path(__file__).resolve().parent.parent


# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/4.0/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = "django-insecure-5))xk@u_hjcu!pd#5sruf5jkt4x#s3-_t^5ou6=1my&x$t*y4t"

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

ALLOWED_HOSTS = []


# Application definition

INSTALLED_APPS = [
    'behave_django',
    "django_auditmatic",
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
]

MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
]

ROOT_URLCONF = "django_model_auditmatic.urls"
AUDITMATIC = {
    "apps": {
        # 'App module name': [model_name, ]
        "auth": {
            "User": {"m2m": any},
        }
    }
}
TENANT_MODEL = "api.tenant.Tenant"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [BASE_DIR / "templates"],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    },
]

WSGI_APPLICATION = "django_model_auditmatic.wsgi.application"


# Database
# https://docs.djangoproject.com/en/4.0/ref/settings/#databases
DATABASES = {
    "default": env.db(
        "DATABASE_URL",
        default="postgres://audit:matic@localhost/auditmatic"
    ),
}


# Password validation
# https://docs.djangoproject.com/en/4.0/ref/settings/#auth-password-validators
_VALIDATOR_PREFIX = "django.contrib.auth.password_validation."
_VALIDATORS = [
    "UserAttributeSimilarityValidator",
    "MinimumLengthValidator",
    "CommonPasswordValidator",
    "NumericPasswordValidator"
]
AUTH_PASSWORD_VALIDATORS = [
    {
        "NAME": f"{_VALIDATOR_PREFIX}{validator}"
    } for validator in _VALIDATORS
]


# Internationalization
# https://docs.djangoproject.com/en/4.0/topics/i18n/

LANGUAGE_CODE = "en-us"

TIME_ZONE = "UTC"

USE_I18N = True

USE_TZ = True


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/4.0/howto/static-files/

STATIC_URL = "static/"

# Default primary key field type
# https://docs.djangoproject.com/en/4.0/ref/settings/#default-auto-field

DEFAULT_AUTO_FIELD = "django.db.models.BigAutoField"



"""
    function to generate the table.
"""


def generate_table(audit_name: str) -> str:
    """
        generates create table sql statement
    :param audit_name:
    :return: create table sql statement
    """
    return f"""
    CREATE TABLE {audit_name}
    (
        change_date timestamptz default now(),
        before      hstore,
        after       hstore
    );
    """



"""
    install audit command
"""
from django.apps import apps

from django_auditmatic.configuration.data_classes import ConfiguredNames
from django_auditmatic.configuration.utils import get_tenant_schemas_and_apps
from django_auditmatic.util import process_model_for_all_schemas


def install_audit():
    """
        installs the auditing triggers and such for the configured models.
    :return:
    """

    # debug = settings.AUDITMATIC.get("debug", False)
    tenant_schemas, schema_apps = get_tenant_schemas_and_apps()

    configured_names = ConfiguredNames.from_settings()

    for model in apps.get_models():
        process_model_for_all_schemas(
            model,
            configured_names,
            schema_apps,
            tenant_schemas,
        )



from os.path import exists

from setuptools import find_packages, setup

setup(
    name="django-model-auditmatic",
    author="Bill Schumacher",
    author_email="william.schumacher@gmail.com",
    packages=find_packages(),
    scripts=[],
    url="https://github.com/BillSchumacher/django-model-auditmatic",
    license="MIT",
    description="Audit support for Django models using PostgreSQL "
    "triggers and stored procedures.",
    long_description=open("README.rst").read() if exists("README.rst") else "",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Framework :: Django",
        "Framework :: Django :: 4.0",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.9",
        "Topic :: Database",
        "Topic :: Software Development :: Libraries",
    ],
    install_requires=["django", "psycopg2"],
    version="0.1.1",
    zip_safe=False,
)



"""
    app config
"""
from django.apps import AppConfig
from django.conf import settings


class DjangoAuditmaticConfig(AppConfig):
    """
    config
    """

    default_auto_field = "django.db.models.BigAutoField"
    name = "django_auditmatic"


if not hasattr(settings, "AUDITMATIC"):
    raise RuntimeError("AUDITMATIC Config not detected.")



"""
    install audit permission
"""
from django.apps import apps
from django.contrib.auth.models import Group

from django_auditmatic.configuration.data_classes import ConfiguredNames
from django_auditmatic.configuration.utils import get_tenant_schemas_and_apps
from django_auditmatic.permission import get_or_create_audit_permission
from django_auditmatic.util import get_model_names


def install_audit_permission() -> None:
    """
    install audit permission for configured models.
    :return: None
    """
    tenant_schemas, schema_apps = get_tenant_schemas_and_apps()
    group, _ = Group.objects.get_or_create(name="Any")
    configured_names = ConfiguredNames.from_settings()
    print("install audit permission")
    for model in apps.get_models():
        model_names = get_model_names(model, configured_names)
        if not model_names:
            continue

        if not len(schema_apps):  # pylint: disable=C1802
            print("install for model", model)
            perm = get_or_create_audit_permission(model)
            if configured_names.allow_any[model._meta.object_name.lower()]:
                group.permissions.add(perm)



"""
when steps
"""
from behave import use_step_matcher, when  # pylint: disable=E0611

from django_auditmatic.install.audit_permission import install_audit_permission
from django_auditmatic.utils.generate import (
    generate_function,
    generate_install_hstore,
    generate_table,
    generate_trigger,
)

use_step_matcher("parse")


@when("the function sql is generated")
def __generate_function_sql(context):
    """
    :type context: behave.runner.Context
    """
    context.statement_generated = generate_function(context.audit_name)


@when("the trigger sql is generated")
def __generate_trigger_sql(context):
    """
    :type context: behave.runner.Context
    """
    context.statement_generated = generate_trigger(
        context.audit_name, context.table_name, "UPDATE"
    )


@when("the table sql is generated")
def __generate_table_sql(context):
    """
    :type context: behave.runner.Context
    """
    context.statement_generated = generate_table(context.audit_name)


@when("the hstore sql is generated")
def step_impl(context):
    """
    :type context: behave.runner.Context
    """
    context.statement_generated = generate_install_hstore()


@when("the permission validator is called with the user model")
def permission_validator_called_with_user(context):
    """
    :type context: behave.runner.Context
    """
    context.result = context.user.has_perm("auth.can_audit_user")


@when("the audit permission is installed")
def audit_permission_installed(context):
    """
    :type context: behave.runner.Context
    """
    install_audit_permission()



"""
    generate functions
"""
from typing import Optional, List

from django_auditmatic.utils.generate.extension import generate_install_hstore
from django_auditmatic.utils.generate.function import generate_function
from django_auditmatic.utils.generate.table import generate_table
from django_auditmatic.utils.generate.trigger import generate_trigger


def generate_sql(
    app_name: str,
    model_name: str,
    schema: str,
    table_name: Optional[str] = None,
    debug: Optional[bool] = True,
    verbs: Optional[List[str]] = None
):
    """
        generates the sql
    :param app_name:
    :param model_name:
    :param schema:
    :param table_name:
    :param debug:
    :param verbs:
    :return:
    """
    if verbs is None:
        verbs = ["INSERT", "UPDATE", "DELETE"]
    table_name, audit_name = get_table_and_audit_name(
        app_name,
        model_name,
        schema,
        table_name
    )
    table_statement = generate_table(audit_name)
    function_statement = generate_function(audit_name)
    triggers = '\n'.join([
        generate_trigger(audit_name, table_name, verb) for verb in verbs
    ])
    statement = combine_statements(table_statement, function_statement, triggers)

    if debug:
        log_generated(statement, model_name, table_name, schema)
    return statement


def get_table_and_audit_name(app_name, model_name, schema, table_name):
    """

    :param app_name:
    :param model_name:
    :param schema:
    :param table_name:
    :return:
    """
    table_name = table_name or f"{app_name}_{model_name}"
    audit_name = f"{schema}.audit_{table_name}"
    table_name = f"{schema}.{table_name}"
    return table_name, audit_name


def combine_statements(table_stmt: str, function_stmt: str, triggers: str) -> str:
    """

    :param table_stmt:
    :param function_stmt:
    :param triggers:
    :return:
    """
    return f"""
    {table_stmt}
    {function_stmt}
    {triggers}
    """


def log_generated(
    statement: str,
    model_name: str,
    table_name: str,
    schema: str
) -> None:
    """

    :param statement:
    :param model_name:
    :param table_name:
    :param schema:
    :return:
    """
    print("Statement generated: ", statement)
    print("Model Name:", model_name)
    print("Table Name:", table_name)
    print("Schema: ", schema)


__all__ = [
    "generate_function",
    "generate_install_hstore",
    "generate_sql",
    "generate_table",
    "generate_trigger",
]



"""
    configuration data classes
"""
from collections import defaultdict
from typing import Dict, List

from django.conf import settings


class AuditMaticAppConfiguration:
    def __init__(self):
        pass


class AuditMaticPermissionConfiguration:
    def __init__(self):
        pass


class AuditMaticConfiguration:
    def __init__(self):
        self.apps = {}
        self.permission = {}


class ConfiguredNames:
    """
    A data object to hold configured name values.
    """

    def __init__(
        self, app_names: List, model_names: Dict, model_m2m_names: Dict, allow_any: Dict
    ):
        self.app_names = app_names
        self.model_names = model_names
        self.allow_any = allow_any
        self.model_m2m_names = model_m2m_names

    @staticmethod
    def process_app_models(
        app_name, app_names, app_models, model_names, model_m2m_names, allow_any
    ):
        """
            process app configured models
        :param allow_any:
        :param app_name:
        :param app_names:
        :param app_models:
        :param model_names:
        :param model_m2m_names:
        :return:
        """
        lowered_app_name = app_name.lower()
        app_names.append(lowered_app_name)
        app_models_type = type(app_models)
        if app_models_type == set:
            names = [(name, {}) for name in app_models]
        elif app_models_type == dict:
            names = app_models.items()
        else:
            raise NotImplementedError()
        for model_name, model_configuration in names:
            lowered_model_name = model_name.lower()
            model_names[lowered_app_name].append(lowered_model_name)
            m2m_key = f"{lowered_app_name}_{lowered_model_name}"
            model_m2m_configured_names = model_configuration.get("m2m", [])

            # print("m2m names ", model_m2m_configured_names)
            if model_m2m_configured_names == any:  # pylint: disable=W0143
                # type(model_m2m_configured_names) == callable and \

                model_m2m_names[m2m_key].append(any)
                # print("is any")
            else:
                for value in model_m2m_configured_names:
                    model_m2m_names[m2m_key].append(value)

            model_allow = model_configuration.get("allow", False)
            if model_allow == any:
                allow_any[lowered_model_name] = True

    @staticmethod
    def from_settings():
        """
            creates ConfiguredNames from settings.
        :return:
        """
        configured_apps = settings.AUDITMATIC["apps"]

        app_names = []
        model_names = defaultdict(list)
        model_m2m_names = defaultdict(list)
        allow_any = defaultdict(bool)
        for app_name, app_models in configured_apps.items():
            ConfiguredNames.process_app_models(
                app_name, app_names, app_models, model_names, model_m2m_names, allow_any
            )

        return ConfiguredNames(app_names, model_names, model_m2m_names, allow_any)


class ModelNames:
    """
    a data object that stores info about a model
    """

    def __init__(self, app_name, model_name, model):
        self.app_name = app_name
        self.model_name = model_name
        self.model = model

    @staticmethod
    def from_model(model):
        """generates an object from a model"""
        app_and_model_name = str(model._meta)
        app_name, model_name = app_and_model_name.split(".")
        return ModelNames(app_name, model_name, model)

    def __repr__(self):
        return f"{self.app_name}_{self.model_name}"



"""
    function to generate the sql function.
"""


def generate_function(audit_name: str) -> str:
    """
        generates sql function for the audit_name
    :param audit_name:
    :return: create function statement
    """
    return f"""
    CREATE OR REPLACE FUNCTION {audit_name}()
        RETURNS TRIGGER
        LANGUAGE plpgsql
    AS $$
    BEGIN
        INSERT INTO {audit_name}(before, after)
            SELECT hstore(old), hstore(new);
        RETURN new;
    END;
    $$;
    """
